################################################################
#
#   Assembly Line - Factory page
#
################################################################

from pygame import Rect, draw
from pygame.locals import *
from albow import Cancel
from albow.widget import Widget
from albow.controls import Column, ImageButton
from albow.palette_view import PaletteView
from albow.resource import get_image
from albow.image_array import get_image_array
from albow.dialogs import alert, ask
from screen import ALScreen
from transformation import Transformation, identity, add, sub
from settings import options
from factory import FactoryError

class FactoryPage(ALScreen):

	bg_color = (128, 180, 204)
	
	def __init__(self, shell):
		ALScreen.__init__(self, shell)
		light = FactoryRunLight(self.game)
		start_button = ImageButton("start_button.png", size = (32, 32), action = self.start_factory)
		stop_button = ImageButton("stop_button.png", size = (32, 32), action = self.stop_factory)
		palette = FactoryPalette(self.game)
		controls = Column([light, start_button, stop_button, palette], spacing = 0)
		controls.right = self.right
		r = Rect(0, 0, controls.left, self.height - 24)
		factory_view = FactoryView(r, self.game)
		self.add(factory_view)
		self.add(controls)
		factory_view.focus()
	
	def start_factory(self):
		try:
			factory = self.game.factory()
			if not factory.running:
				factory.check_ready()
				unpurchased = factory.unpurchased_machines()
				if unpurchased:
					self.purchase_machines(unpurchased)
				self.game.factory().start()
		except Cancel:
			pass
		except FactoryError, e:
			alert(str(e))
	
	def purchase_machines(self, machines):
		if not options.dbg_no_costs:
			self.pay_for_machines(machines)
		for machine in machines:
			machine.purchased = True
	
	def pay_for_machines(self, machines):
		cost = 0
		for machine in machines:
			cost += machine.cost
		company = self.game.company
		cash = company.cash
		if cost > cash:
			alert("Not enough cash to purchase machinery.\n"
				"Cost: $%s Available funds: $%s" % (cost, cash))
			raise Cancel
		if ask("Purchase new machinery for $%s?" % cost) <> "OK":
			raise Cancel
		company.spend_cash("Machinery purchase", cost)

	def stop_factory(self):
		self.game.factory().stop()
	
#-----------------------------------------------------------------------

class FactoryRunLight(Widget):

	images = get_image_array("factory_run_light.png", 2, border = 2)
	
	def __init__(self, game):
		self.game = game
		r = self.images[0].get_rect()
		Widget.__init__(self, r)
	
	def draw(self, surf):
		image = self.images[self.game.factory().running]
		surf.blit(image, (0, 0))

#-----------------------------------------------------------------------

class FactoryPalette(PaletteView):

	cell_size = (32, 32)
	tools = ['move', 'add', 'remove', 'configure', 'inspect']
	highlight_style = 'fill'

	def __init__(self, game):
		self.game = game
		PaletteView.__init__(self, self.cell_size, self.num_items(), 1)
	
	def num_items(self):
		return len(self.tools)
	
	def item_is_selected(self, n):
		return self.game.factory_tool == self.tools[n]
	
	def click_item(self, n, e):
		self.game.factory_tool = self.tools[n]
	
	def draw_item(self, surf, n, rect):
		image_name = self.tools[n] + "_tool.png"
		image = get_image(image_name)
		r = image.get_rect()
		r.center = rect.center
		surf.blit(image, r)

#-----------------------------------------------------------------------

class FactoryView(Widget):

	error_marker_image = get_image("factory_error_marker.png")

	gridline_color = (96, 140, 159)
	mouse_position = (0, 0)
	moving_machine = None
	moving_delta = (0, 0)
	moving_rotation = 0
	feedback_okay = False
	feedback_rects = None
	
	def __init__(self, rect, game):
		Widget.__init__(self, rect)
		self.game = game
	
	def factory(self):
		return self.game.factory()
	
	def factory_running(self):
		return self.factory().running
	
	def view_to_global_point(self, p):
		return p

	def global_to_view_point(self, p):
		return p

	def draw(self, surf):
		factory = self.factory()
		if factory:
			self.draw_factory(surf, factory)
	
	def draw_factory(self, surf, factory):
		self.draw_grid_lines(surf, factory)
		factory.draw(surf, identity)
		self.draw_feedback(surf, factory)
		self.draw_error(surf, factory)
	
	def draw_grid_lines(self, surf, factory):
		cw, ch = factory.grid_size
		wc, hc = factory.cell_size
		color = self.gridline_color
		w = cw * wc
		h = ch * hc
		for cx in xrange(cw + 1):
			x = cx * wc
			draw.line(surf, color, (x, 0), (x, h))
		for cy in xrange(ch + 1):
			y = cy * hc
			draw.line(surf, color, (0, y), (w, y))

	def draw_feedback(self, surf, factory):
		rects = None
		tool = self.game.factory_tool
		if tool == 'add' or tool == 'move':
			factory = self.factory()
			if not factory.running:
				okay = self.feedback_okay
				rects = self.feedback_rects
		if rects:
			if okay:
				color = (0, 255, 0)
			else:
				color = (255, 0, 0)
			for rl in rects:
				for r in rl:
					draw.rect(surf, color, r, 1)
	
	def draw_error(self, surf, factory):
		p = factory.error_position
		if p:
			q = self.global_to_view_point(p)
			image = self.error_marker_image
			r = image.get_rect()
			r.center = q
			surf.blit(image, r)
	
	def machine_to_add(self):
		mtype = self.game.selected_machine_type
		if mtype:
			machine = self.game.company.get_machine_from_storage(mtype)
			if not machine:
				machine = mtype()
			return machine

	def mouse_move(self, e):
		#print "FactoryPage.mouse_move" ###
		tool = self.game.factory_tool
		if tool == 'add':
			self.add_tool_move(e)
	
	def mouse_down(self, e):
		tool = self.game.factory_tool
		if tool == 'move':
			self.move_tool_click(e)
		elif tool == 'add':
			self.add_tool_click(e)
		elif tool == 'remove':
			self.remove_tool_click(e)
		elif tool == 'configure':
			self.configure_tool_click(e)
		elif tool == 'inspect':
			self.inspect_tool_click(e)
	
	def mouse_drag(self, e):
		tool = self.game.factory_tool
		if tool == 'move':
			self.move_tool_drag(e)
	
	def mouse_up(self, e):
		tool = self.game.factory_tool
		if tool == 'move':
			self.move_tool_release(e)
	
	def key_down(self, e):
		if e.key == K_r:
			self.rotate_cmd(e)
		else:
			Widget.key_down(self, e)
	
	def add_tool_move(self, e):
		factory = self.factory()
		if not factory.running:
			mtype = self.game.selected_machine_type
			if mtype:
				self.moving_mouse_position = self.view_to_global_point(e.local)
				self.moving_delta = (0, 0)
				self.moving_machine_type = mtype
				self.update_moving_feedback()
				#p = self.view_to_global_point(e.local)
				#t = Transformation(self.moving_rotation, p)
				#okay, rects, t = factory.new_machine_type_position(mtype, t)
				#self.feedback_okay = okay
				#self.feedback_rects = rects
				self.invalidate()
				return
		if self.feedback_rects:
			self.feedback_rects = None
			self.invalidate()
	
	def add_tool_click(self, e):
		factory = self.factory()
		if not factory.running:
			machine = self.machine_to_add()
			if machine:
				self.moving_mouse_position = self.view_to_global_point(e.local)
				t = self.moving_transformation()
				if factory.add_machine(machine, t):
					self.cancel_moving_feedback()
				else:
					if machine.purchased:
						self.game.company.put_machine.into_storage(machine)

	def move_tool_click(self, e):
		factory = self.factory()
		if not factory.running:
			p = self.view_to_global_point(e.local)
			machine = factory.find_machine_at(p)
			if machine:
				self.moving_machine = machine
				self.moving_machine_type = machine.__class__
				self.moving_mouse_position = p
				self.moving_delta = sub(machine.translation, p)
				self.moving_rotation = machine.rotation
				self.update_moving_feedback()
	
	def move_tool_drag(self, e):
		machine = self.moving_machine
		if machine:
			self.moving_mouse_position = self.view_to_global_point(e.local)
			self.update_moving_feedback()
	
	def update_moving_feedback(self):
		mtype = self.moving_machine_type
		if mtype:
			machine = self.moving_machine
			t = self.moving_transformation()
			okay, rects, t = \
				self.factory().new_machine_type_position(mtype, t, ignore_machine = machine)
			self.feedback_okay = okay
			self.feedback_rects = rects
			self.invalidate()
	
	def cancel_moving_feedback(self):
		self.moving_machine = None
		self.moving_machine_type = None
		self.feedback_rects = None
	
	def move_tool_release(self, e):
		machine = self.moving_machine
		if machine:
			if self.feedback_okay:
				factory = self.factory()
				factory.remove_machine(machine)
				old_t = machine.local_to_global_transformation()
				t = self.moving_transformation()
				if not factory.add_machine(machine, t):
					if not factory.add_machine(machine, old_t):
						# Shouldn't be possible, but just in case
						self.game.company.put_machine_into_storage(machine)
			self.cancel_moving_feedback()
	
	def moving_transformation(self):
		p = add(self.moving_delta, self.moving_mouse_position)
		r = self.moving_rotation & self.moving_machine_type.rotation_mask
		return Transformation(r, p)
	
	def remove_tool_click(self, e):
		factory = self.factory()
		if not factory.running:
			#p = self.view_to_global_point(e.local)
			#machine = factory.find_machine_at(p)
			machine = self.find_clicked_machine(e)
			if machine:
				factory.remove_machine(machine)
				if machine.purchased:
					self.game.company.put_machine_into_storage(machine)
	
	def configure_tool_click(self, e):
		if not self.factory_running():
			machine = self.find_clicked_machine(e)
			if machine:
				editor_class = machine.editor_class
				if editor_class:
					editor = editor_class(machine)
					editor.present()
	
	def inspect_tool_click(self, e):
		machine = self.find_clicked_machine(e)
		if machine:
			self.factory().inspect(machine)
			self.invalidate()
	
	def find_clicked_machine(self, e):
		p = self.view_to_global_point(e.local)
		return self.factory().find_machine_at(p)

	def rotate_cmd(self, e):
		tool = self.game.factory_tool
		if tool == 'add' or tool == 'move':
			self.change_rotation()
		#elif tool == 'move':
		#	machine = self.moving_machine
		#	if machine:
		#		self.change_rotation(machine.rotation_mask)
	
	def change_rotation(self):
		self.moving_rotation = (self.moving_rotation + 1) % 4
		self.update_moving_feedback()

	#def change_rotation(self, mask):
	#	if mask:
	#		self.moving_rotation = ((self.moving_rotation + 1) % 4) & mask
	#		self.update_moving_feedback()
