################################################################
#
#   Assembly Line - Machine
#
################################################################

from transformation import Transformation, add, rotate_vector

class MachineNotConfiguredError(Exception):
	pass

#----------------------------------------------------------------------

class MachineType(type):
	#  title                 string
	#  local_bounding_rect   Rect
	#  local_covering_rects  list[layer] of [Rect]

	def bounding_rect(self, t):
		lr = self.local_bounding_rect
		gr = t.rect(lr)
		#print "MachineType.global_bounding_rect:" ###
		#print "rotation =", t.rotation, "translation =", t.translation ###
		#print "lr =", lr, "gr =", gr ###
		return gr
	
	def covered_rects(self, t):
		return [
			[t.rect(r) for r in rects]
				for rects in self.local_covered_rects]

#----------------------------------------------------------------------

class Machine(object):
	#  translation           vector
	#  rotation              int
	#  purchased             boolean
	#
	#  Class attributes:
	#
	#  rotation_mask         int
	#  on_left               boolean
	#  layer                 int
	#  cost                  int
	#  is_conveyor           boolean
	#  is_conveyor_end       boolean
	#  is_conveyor_drive     boolean
	#  is_inspectable        boolean
	#  editor_class          Widget class
	
	__metaclass__ = MachineType

	rotatable = False
	on_left = False
	is_conveyor = False
	is_conveyor_end = False
	is_conveyor_drive = False
	is_inspectable = False
	editor_class = None
	rotation_mask = 0
	
	translation = (0, 0)
	rotation = 0
	purchased = False
	
	def local_to_global_transformation(self):
		return Transformation(self.rotation, self.translation)

	def global_bounding_rect(self):
		t = self.local_to_global_transformation()
		return self.__class__.bounding_rect(t)
	
	def machine_bounding_rect(self, t):
		return self.__class__.bounding_rect(t.trans(self))

	def global_covered_rects(self):
		t = self.local_to_global_transformation()
		return self.__class__.covered_rects(t)
	
	def global_vector(self, v):
		return rotate_vector(self.rotation, v)
	
	def global_origin(self):
		return self.translation
	
	def global_point(self, p):
		t = self.local_to_global_transformation()
		return t.point(p)

	def global_rect(self, r):
		t = self.local_to_global_transformation()
		return t.rect(r)
	
	def position_global(self, p):
		self.translation = p
	
	def set_transformation(self, t):
		self.rotation = t.rotation
		self.translation = t.translation
	
	def offset_global(self, d):
		ld = rotate_vector((-self.rotation) % 4, d)
		self.translation = add(ld, self.translation)

	def local_topleft(self):
		return self.local_bounding_rect.topleft
	
	def in_layer(self, layer):
		return self.layer == layer
	
	def draw(self, surf, t, layer):
		self.draw_local_layer(surf, t.trans(self), layer)
	
	def draw_local_layer(self, surf, t, layer):
		self.draw_local(surf, t)

	def begin_frame(self, factory):
		pass

	def check_ready(self):
		pass

	def handle_part(self, factory, part):
		pass

	def update_neighbourhood(self, factory):
		pass

	def part_removed(self, part):
		pass

	def removed_from_factory(self):
		pass

	def find_unpurchased_components(self, result):
		pass

	def part_in_rect(self, part, local_rect):
		r = self.global_rect(local_rect)
		return r.collidepoint(part.position)

	def move_part_to(self, factory, part, local_point):
		p = self.global_point(local_point)
		factory.move_part_to(part, p)
