################################################################
#
#   Assembly Line - Quality control Station
#
################################################################

from pygame import Rect
from albow.resource import get_image, get_font
from albow.image_array import get_image_array
from albow.controls import Label, Grid
from machine import Machine, MachineNotConfiguredError
from part import compare_parts
from game import game
from controls import MachineEditor, ProductKeyField

#------------------------------------------------------------------------

class QCStationEditor(MachineEditor):

	def create_contents(self, machine):
		return Grid([
			[Label("Compare with Product"),
				ProductKeyField(attribute = 'product_key', model = machine)],
		])
		
#------------------------------------------------------------------------

class QCStation(Machine):
	#  product_key   string

	title = "QC Station"
	cost = 500
	layer = 1
	icon = get_image("qc_station_icon.png")
	image0 = get_image("qc_station_table.png")
	image1 = get_image("qc_station.png")
	lights = get_image_array("qc_station_lights.png", 2, border = 2)
	editor_class = QCStationEditor
	is_inspectable = True

	local_bounding_rect = Rect(-32, -96, 96, 96)
	local_handling_rect = Rect(-18, -32, 4, 64)
	local_light_point = (-16, -60)
	local_rejection_point = (32, -64)
	local_discard_point = (96, -64)
	
	local_covered_rects = [
		[Rect(-32, -96, 96, 64)],
		[Rect(-32, -32, 32, 64), Rect(64, -96, 32, 64)]
	]
	
	checking_frames = 4
	
	product_key = None
	tolerance = 0.0
	checking_part = None
	last_checked_part = None
	last_rejected_part = None
	#state = 'idle' # or 'checking'
	timeout = 0
	last_result = None # or 'passed' or 'failed'
	
	def removed_from_factory(self):
		self.checking_part = None
		self.last_checked_part = None
		self.last_rejected_part = None
		self.last_result = None
		self.state = 'idle'
		self.timeout = 0
	
	def check_ready(self):
		if not self.product_key:
			raise MachineNotConfiguredError
	
	def in_layer(self, layer):
		return True

	def draw_local_layer(self, surf, t, layer):
		if layer == 0:
			self.draw_layer0(surf, t)
		else:
			self.draw_layer1(surf, t)
	
	def draw_layer0(self, surf, t):
		t.draw_image(surf, [self.image0], (0, -96))
	
	def draw_layer1(self, surf, t):
		org = self.local_topleft()
		t.draw_image(surf, [self.image1], (-32, -64))
		lr = self.last_result
		if lr is not None:
			light = self.lights[lr == 'passed']
			t.draw_image(surf, [light], self.local_light_point)

	def handle_part(self, factory, part):
		#print "QCStation: Considering part:", part ###
		if part is not self.last_checked_part:
			if self.part_in_rect(part, self.local_handling_rect):
				if not self.checking_part:
					self.start_checking_part(factory, part)
					return True
				else:
					self.reject_part(factory, self.checking_part)
					self.reject_part(factory, part)
					self.state = 'idle'
				
	def start_checking_part(self, factory, part):
		#print "QCStation: Starting to check part:", part ###
		assert not self.checking_part
		self.checking_part = part
		part.handler = self
		self.last_checked_part = part
		self.timeout = self.checking_frames
	
	def accept_part(self, factory):
		#print "QCStation: Accepting part" ###
		self.stop_checking_part(factory)
		self.last_result = 'passed'
	
	def reject_part(self, factory, part):
		#print "QCStation: Rejecting part:", part ###
		self.stop_checking_part(factory)
		self.discard_last_rejected_part(factory)
		self.move_part_to(factory, part, self.local_rejection_point)
		self.last_rejected_part = part
		self.last_result = 'failed'
	
	def stop_checking_part(self, factory):
		part = self.checking_part
		if part:
			#print "QCStation: Stopping checking of part:", part ###
			part.handler = None
			self.checking_part = None
			self.timeout = 0
	
	def discard_last_rejected_part(self, factory):
		part = self.last_rejected_part
		if part:
			#print "QCStation: Discarding last rejected part:", part ###
			self.move_part_to(factory, part, self.local_discard_point)
			part.handler = None
			self.last_rejected_part = None

	def begin_frame(self, factory):
		#print "QCStation.begin_frame: Checking", self.checking_part, "timeout =", self.timeout ###
		if self.timeout > 0:
			self.timeout -= 1
		else:
			if self.checking_part:
				self.perform_check(factory)

	def perform_check(self, factory):
		#print "QCStation: Performing check of", self.checking_part ###
		part = self.checking_part
		product = game.product_with_key(self.product_key)
		if part and product:
			error = compare_parts(part, product.part)
			#print "QCStation: Part error =", error ###
			if error <= self.tolerance:
				self.accept_part(factory)
			else:
				self.reject_part(factory, part)
